@@ -1,12 +1,12 @@
 # $Id$
 
 package TheSchwartz;
-use 5.008;
+use 5.008001;
 use strict;
 use fields
-    qw( databases retry_seconds dead_dsns retry_at funcmap_cache verbose all_abilities current_abilities current_job cached_drivers driver_cache_expiration scoreboard prioritize floor batch_size );
+    qw( databases retry_seconds dead_dsns retry_at funcmap_cache verbose all_abilities current_abilities current_job cached_drivers driver_cache_expiration scoreboard prioritize floor batch_size strict_remove_ability);
 
-our $VERSION = "1.11";
+our $VERSION = "1.13";
 
 use Carp qw( croak );
 use Data::ObjectDriver::Errors;
@@ -45,6 +45,8 @@ sub new {
         || 0;
     $client->{batch_size} = delete $args{batch_size} || $FIND_JOB_BATCH_SIZE;
 
+    $client->{strict_remove_ability} = delete $args{strict_remove_ability};
+
     my $floor = delete $args{floor};
     $client->set_floor($floor) if ($floor);
 
@@ -134,6 +136,7 @@ sub mark_database_as_dead {
     my ($hashdsn) = @_;
     $client->{dead_dsns}{$hashdsn} = 1;
     $client->{retry_at}{$hashdsn}  = time + $client->{retry_seconds};
+    $client->debug("Disabling DB $hashdsn because " . ($client->driver_for($hashdsn)->last_error() || 'unknown'));
 }
 
 sub is_database_dead {
@@ -417,8 +420,16 @@ JOB:
             $server_time + ( $worker_class->grab_for || 1 ) );
 
         ## Update the job in the database, and end the transaction.
-        if ( $driver->update( $job, { grabbed_until => $old_grabbed_until } )
-            < 1 )
+        ## NOTE: For some reason, D::OD doesn't ensure the object's value is
+        ##       in bounds of original search query. so we need to be more paranoic
+        ##       to make sure it's not grabbed by other workers.
+        my $unixtime = $driver->dbd->sql_for_unixtime;
+        if ( $driver->update( $job, {
+            grabbed_until => [
+                '-and',
+                { op => '=', value => $old_grabbed_until},
+                \" <= $unixtime"
+            ]}) < 1 )
         {
             ## We lost the race to get this particular job--another worker must
             ## have got it and already updated it. Move on to the next job.
@@ -636,7 +647,8 @@ sub work_once {
 
     ## If we didn't find anything, restore our full abilities, and try
     ## again.
-    if ( !$job
+    if (   !$job
+        && !$client->{strict_remove_ability}
         && @{ $client->{current_abilities} } < @{ $client->{all_abilities} } )
     {
         $client->restore_full_abilities;
@@ -660,7 +672,8 @@ sub work_once {
     ## from our list of current abilities. So the next time we look for a
     ## we'll find a job for a different funcname. This prevents starvation of
     ## high funcid values because of the way MySQL's indexes work.
-    $client->temporarily_remove_ability($class);
+## BUGBUG this looks odd since ordering by job_id should limit any skew ...
+    $client->temporarily_remove_ability($class) unless($client->{strict_remove_ability});
 
     $class->work_safely($job);
 
@@ -875,6 +888,16 @@ sub set_current_job {
     $client->{current_job} = shift;
 }
 
+sub strict_remove_ability {
+    my TheSchwartz $client = shift;
+    return $client->{strict_remove_ability};
+}
+
+sub set_strict_remove_ability {
+    my TheSchwartz $client = shift;
+    $client->{strict_remove_ability} = shift;
+}
+
 DESTROY {
     foreach my $arg (@_) {
 
@@ -1030,6 +1053,12 @@ The number of seconds after which to try reconnecting to apparently dead
 databases. If not given, TheSchwartz will retry connecting to databases after
 30 seconds.
 
+=item * C<strict_remove_ability>
+
+By default when work_once does not find a job it will reset current_abilities to
+all_abilities and look for a job. Setting this option will prevent work_once from
+resetting abilities if it can't find a job for the current capabilities.
+
 =back
 
 =head2 C<$client-E<gt>list_jobs( %args )>
@@ -1117,6 +1146,10 @@ Set the C<floor<gt> value as described in the constructor.
 
 Set the C<batch_size<gt> value as described in the constructor.
 
+=head2 C<$client-E<gt>set_strict_remove_ability( $strict_remove_ability )>
+
+Set the C<strict_remove_ability<gt> value as described in the constructor.
+
 =head1 WORKING
 
 The methods of TheSchwartz clients for use in worker processes are:
